package org.apidesign.apifest08.currency;

import java.util.Currency;
import java.util.Hashtable;

import org.apidesign.apifest08.currency.exceptions.ConvertorException;
import org.apidesign.apifest08.currency.exceptions.InvalidCurrencyException;
import org.apidesign.apifest08.currency.exceptions.UnknownConvertorException;

/**
 * This is the skeleton class for your API. You need to make it public, so it is accessible to your client code
 * (currently in Task1Test.java) file.
 * <p>
 * Feel free to create additional classes or rename this one, just keep all the API and its implementation in this
 * package. Do not spread it outside to other packages.
 */
public class Convertor {

  private static Hashtable<String, ExchangeRate> exchangeRates;

  private ExchangeRate exchangeRate12;
  private ExchangeRate exchangeRate21;

  private Convertor(Currency currency1, Currency currency2) throws UnknownConvertorException {
    String key12 = currency1.getCurrencyCode() + currency2.getCurrencyCode();
    String key21 = currency2.getCurrencyCode() + currency1.getCurrencyCode();

    if (!(exchangeRates.containsKey(key12) && exchangeRates.containsKey(key21))) {
      throw new UnknownConvertorException("Selected convertor (" + currency1.getCurrencyCode() + "->"
          + currency2.getCurrencyCode() + ") has not defined any rates!!!");
    }

    this.exchangeRate12 = exchangeRates.get(key12);
    this.exchangeRate21 = exchangeRates.get(key21);
  }

  /**
   * Sets convertor rate for selected currencies.
   * @param currency1
   *          one of the currencies we want to convert to/from
   * @param currency2
   *          the other currency
   * @param rate
   *          exchange rate from currency1 to currency2
   * @param unit
   *          unit of exchangeRate (USD->CZK - unit=1, you exchange one dollar, SKK->CZK unit=100, exchange rate is for
   *          100SKK)
   */
  public static void setConvertorRates(Currency currency1, Currency currency2, double rate, double unit) {
    if (currency1 == null || currency2 == null) {
      throw new ConvertorException("None of the currencies should be null!!!");
    }

    if (rate <= 0 || unit <= 0) {
      throw new ConvertorException("Rate(" + rate + ") and unit(" + unit + ") has to be grater then zero!!!");
    }

    if (exchangeRates == null) {
      exchangeRates = new Hashtable<String, ExchangeRate>();
    }

    String key12 = currency1.getCurrencyCode() + currency2.getCurrencyCode();
    String key21 = currency2.getCurrencyCode() + currency1.getCurrencyCode();
    double recountedRate = (unit / rate) * unit;

    exchangeRates.put(key12, new ExchangeRate(currency1, currency2, rate, unit));
    exchangeRates.put(key21, new ExchangeRate(currency2, currency1, recountedRate, unit));

  }

  /**
   * Creates new instance of convertor.
   * @param currency1
   *          one of the currencies we want to convert to/from
   * @param currency2
   *          the other currency
   * @return new instance of convertor
   * @throws UnknownConvertorException
   *           thrown if convertor for selected currencies has not been defined
   */
  public static Convertor getConvertorInstance(Currency currency1, Currency currency2) throws UnknownConvertorException {
    if (currency1 == null || currency2 == null) {
      throw new ConvertorException("None of the currencies should be null!!!");
    }
    return new Convertor(currency1, currency2);
  }

  /**
   * Converts selected amout of selected currency to other currency of this convertor instance.
   * @param amount
   *          amount to convert
   * @param originalCurrency
   *          currency of this amount
   * @param newCurrency
   *          currency to which we want convert
   * @return converted amount
   * @throws InvalidCurrencyException
   *           while one or both currencies doesn't fit for this convertor
   */
  public double convert(double amount, Currency originalCurrency, Currency newCurrency) throws InvalidCurrencyException {
    ExchangeRate actualyUsedExchangeRate = null;

    if (originalCurrency == null) {
      throw new ConvertorException("Original currency is null!!!");
    }

    if (newCurrency == null) {
      throw new ConvertorException("Destination currency is null!!!");
    }

    actualyUsedExchangeRate = getExchangeRate(originalCurrency, newCurrency);

    return countResult(actualyUsedExchangeRate, amount);
  }

  private double countResult(ExchangeRate actualyUsedExchangeRate, double amount) {
    return amount * actualyUsedExchangeRate.getRate() / actualyUsedExchangeRate.getUnit();
  }

  /**
   * Decides the direction of conversion and returns instance of actual exchange rate.
   * @param actualCurrency
   *          actual currency we want to convert
   * @return actual exchange rate of this convertor for selected currency
   */
  private ExchangeRate getExchangeRate(Currency originalCurrency, Currency newCurrency) throws InvalidCurrencyException {
    ExchangeRate actualyUsedExchangeRate = null;

    if (originalCurrency.getCurrencyCode().equals(exchangeRate12.getOriginalCurrency().getCurrencyCode())
        && newCurrency.getCurrencyCode().equals(exchangeRate12.getNewCurrency().getCurrencyCode())) {
      actualyUsedExchangeRate = exchangeRate12;
    } else if (originalCurrency.getCurrencyCode().equals(exchangeRate21.getOriginalCurrency().getCurrencyCode())
        && newCurrency.getCurrencyCode().equals(exchangeRate21.getNewCurrency().getCurrencyCode())) {
      actualyUsedExchangeRate = exchangeRate21;
    } else {
      throw new InvalidCurrencyException("This convertor " + this
          + " could not be used for converting selected currencies (" + originalCurrency.getCurrencyCode() + "->"
          + newCurrency.getCurrencyCode() + ") !!!");
    }

    return actualyUsedExchangeRate;
  }

  public String toString() {
    String currency1Code = exchangeRate12.getOriginalCurrency().getCurrencyCode();
    String currency2Code = exchangeRate12.getNewCurrency().getCurrencyCode();
    return "Converter [" + currency1Code + "->" + currency2Code + ", " + currency2Code + "->" + currency1Code + "]";
  }
}
