package org.apidesign.apifest08.currency;

import java.math.BigDecimal;
import java.util.Currency;

/**
 * An amount of a currency.
 * Immutable.
 * @author jdvorak
 */
public class MonetaryAmount {

    private final BigDecimal amount;
    private final Currency currency;
    
    /**
     * A new amount.
     * @param amount the quantity of the currency; must not be null
     * @param currency the currency; must not be null
     */
    public MonetaryAmount( final BigDecimal amount, final Currency currency ) {
        this.amount = amount;
        this.currency = currency;
        if ( amount == null ) {
            throw new NullPointerException( "The amount" );
        }
        if ( currency == null ) {
            throw new NullPointerException( "The currency" );
        }
    }
    
    /**
     * A new amount.
     * @param amount the quantity of the currency; must not be null
     * @param currency the currency; must not be null
     */
    public MonetaryAmount( final double amount, final Currency currency ) {
        this( new BigDecimal( amount ), currency );
    }
    
    /**
     * The amount.
     * @return the amount
     */
    public BigDecimal getAmount() {
        return amount;
    }
    
    /**
     * The currency.
     * @return the currency
     */
    public Currency getCurrency() {
        return currency;
    }

    /**
     * The string representation of the monetary amount.
     * @return the amount, a non-breakable space, the currency
     */
    @Override
    public String toString() {
        return amount.toPlainString() + "\u00a0" + currency.toString();
    }
    
    /**
     * Two monetary amounts are equal to each other iff they have equal amounts of equal currencies.
     * @param other the other object
     * @return equality
     */
    @Override
    public boolean equals( final Object other ) {
        if ( other instanceof MonetaryAmount ) {
            final MonetaryAmount otherMonetaryAmount = (MonetaryAmount) other;
            return getAmount().equals( otherMonetaryAmount.getAmount() ) && getCurrency().equals( otherMonetaryAmount.getCurrency() );
        }
        return false;
    }
    
    /**
     * The hash code combines the hash codes of the amount and of the currency.
     * @return hash code
     */
    @Override
    public int hashCode() {
        return amount.hashCode() * 37 + currency.hashCode();
    }
    
}
