package org.apidesign.apifest08.currency;

import java.math.BigDecimal;
import java.math.MathContext;
import java.math.RoundingMode;

/** Convertor able to convert amount from one currency to other currency.
 * <p>
 * Conversion method are:
 * <ul>
 * <li>{@link #convert(ConvertorCurrency, ConvertorCurrency, BigDecimal)} - convert
 * using exchange rate specified in exchange rate provide. This method is not try 
 * use reverted excahnage rate. 
 * <li>{@link #convertWithReversibleRates(ConvertorCurrency, ConvertorCurrency, BigDecimal)} - 
 * convert using exchange rate specified in exchange rate provide. This method
 * is not trying to use reverted exchange rate.
 * </ul>
 * 
 * Exchange rate is provided by {@link ExchangeRateProvider}.
 */
public class Convertor {
    private Convertor[] convertors;

    /** Create new <code>Convertor</code> as merge of provided convertors. Merged convertor will use
     * provided convertors to convert between currencies.
     * <p>
     * Only one should be able to provide conversion between currencies. If more than one convertos
     * are able to convert currency, one of conversions will be used (it is not defined which).
     * 
     * @since version 2
     * @param convertors Convertor used to create merge-convertor.
     * @return Returns new convertor instance.
     */
    public static Convertor createConvertorAsMerge(Convertor[] convertors) {
        return new Convertor(convertors);
    }
    
    boolean remainderAllowed = true; //if false, remained is not allowed (should be true ideally, but can't handle it now)
    ExchangeRateProvider exchangeRateProvider; 
    
    /** Create simle convertor.
     */
    private Convertor() {
        this.convertors=new Convertor[0];
    }
    
    /** Create merge convertor.
     */            
    private Convertor(Convertor[] convertors) {
        this.convertors = convertors;       
    }
    
    /**
     * Create new <code>Convertor</code> using <code>ExchangeRateProvider</code>.
     * 
     * @param exchangeRateProvider {@link ExchangeRateProvider} used to get exchange rate.
     * 
     * @return Returns <code>Convertor</code> which can be used to convert money.
     * @since version1
     */
    public static Convertor createConvertor(ExchangeRateProvider exchangeRateProvider) {
        Convertor c = new Convertor();

        c.exchangeRateProvider = exchangeRateProvider;
        return c;
    }
    
    /**
     * Convert <code>amount</code> from <code>fromCurrency</code> to <code>toCurrency</code> as specified
     * in <code>ExchangeRateProvider</code>.
     * 
     * @param amount Amount which should be converted. Can't be negative value (can be zero or positive).
     * @return Return <code>ConversionResult</code> which holds conversion result.
     * @since version1
     * @deprecated since version2. Use {@link #convert(ConvertorCurrency, ConvertorCurrency, BigDecimal) } - explicitly specify conversion currencies.
     */
    public ConversionResult convert(BigDecimal amount) {
        return convertValue(exchangeRateProvider.getFromCurrency(), exchangeRateProvider.getToCurrency(),amount, false,false);
    }
    
    /**
     * Convert <code>amount</code> from <code>toCurrency</code> to <code>fromCurrency</code> as specified
     * in <code>ExchangeRateProvider</code>. This is <em>reverted</em> order than suggested by  names of currency fields in <code>ExchangeRate</code>.
     * 
     * @param amount Amount which should be converted. Can't be negative value (can be zero or positive).
     * @return Return <code>ConversionResult</code> which holds conversion result.
     * @since version1
     * @deprecated since version2. Use {@link #convert(ConvertorCurrency, ConvertorCurrency, BigDecimal) } - explicitly specify conversion currencies.
     */
    public ConversionResult convertBack(BigDecimal amount) {
        return convertValue(exchangeRateProvider.getFromCurrency(), exchangeRateProvider.getToCurrency(),amount, true,false);
    }

    private ConversionResult convertUsingSimpleConvertor(ConvertorCurrency fromCurrency, ConvertorCurrency toCurrency, boolean reversibleExRate, BigDecimal amount, boolean convertBack) throws ConversionNotSupportedException, RuntimeException {
        ConversionResult result = new ConversionResult();

        //ExchangeRate rate = exchangeRateProvider.getExchangeRate();
        ExchangeRate rate;
        if (reversibleExRate) {
            rate = exchangeRateProvider.getReversibleExchangeRate(fromCurrency, toCurrency);
        } else {
            rate = exchangeRateProvider.getExchangeRate(fromCurrency, toCurrency);
        }
        if (rate == null) {
            return null;
        }
        
        int fromFranctionDigits = fromCurrency.getDefaultFractionDigits();
        int toFractionDigits = toCurrency.getDefaultFractionDigits();

        if (toFractionDigits != 2) {
            throw new ConvertorException("Can't process currency with defaultFractionDigits!=2, " + exchangeRateProvider.getToCurrency() + " has " + toFractionDigits + " defaultFractionDigits");
        }
        if (fromFranctionDigits != 2) {
            throw new ConvertorException("Can't process currency with defaultFractionDigits!=2, " + exchangeRateProvider.getFromCurrency() + " has " + fromFranctionDigits + " defaultFractionDigits");
        }

        if (amount.signum() == -1) {
            throw new RuntimeException("Can convert only non-negative value, current value is " + amount);
        }


        MathContext context = new MathContext(0, RoundingMode.DOWN);

        BigDecimal from;
        BigDecimal to;
        if (convertBack) {
            //converting in reverted way
            to = rate.getFromValue();
            from = rate.getToValue();
        } else {
            //converting in normal way
            from = rate.getFromValue();
            to = rate.getToValue();
        }

        BigDecimal amountCent = amount.movePointRight(toFractionDigits);

        final BigDecimal multiplied = amountCent.multiply(to, context);
        BigDecimal[] division = multiplied.divideAndRemainder(from, context);

        if (!remainderAllowed && !(BigDecimal.ZERO.equals(division[1]))) {
            throw new RuntimeException("Remained is not allowed - remaining amount is " + division[1] + " cents");
        } else {
            result.setRemainder(BigDecimal.ZERO);
        }

        BigDecimal converted = division[0].movePointLeft(toFractionDigits);
        converted = converted.setScale(toFractionDigits); //XXX ugly
        result.setConverted(converted);
        //result.setRemainder(...);
        return result;
    }
    

    private ConversionResult convertValue(ConvertorCurrency fromCurrency, ConvertorCurrency toCurrency,BigDecimal amount, boolean convertBack,boolean reversibleExRate) throws RuntimeException {
        //result.setRemainder(...);
        if (convertors.length==0) {
            return convertUsingSimpleConvertor(fromCurrency, toCurrency, reversibleExRate, amount, convertBack);
        } else {
            ConversionResult result = null;
            for (int i = 0;i<convertors.length;i++) {
                Convertor subConvertor = convertors[i];
                result = subConvertor.convertValue(fromCurrency, toCurrency, amount, convertBack, reversibleExRate);
                if (result!=null) {
                    break;
                }
            }
            return result;
        }
    }
    
    /**
     * Convert <code>value</code> from <code>fromCurrency</code> to <code>toCurrency</code>.
     * <p>
     * Exchange rate is provided by exchange rate provider which was specified when Convertor was created.
     * This method is using only exchange rate from->to and not trying to use reverted excange rate to->from.
     * 
     * @param fromCurrency Source currency to convert from.
     * @param toCurrency Target currency to convert to.
     * @param value Value in source currency which should be converted.
     * @return Return conversion result.
     * @since version2
     * @throws ConversionNotSupportedException If conversion from <code>fromCurrency</code> to <code>toCurrency</code> is not supported.
     */
    public ConversionResult convert(ConvertorCurrency fromCurrency, ConvertorCurrency toCurrency, BigDecimal value) {
        ConversionResult result = convertValue(fromCurrency, toCurrency, value, false,false);
        if (result==null) {
            //throw new ConversionNotSupportedException("Conversion from " + fromCurrency + " to " + toCurrency + " is not supported");
            throw new ConversionNotSupportedException(fromCurrency.getCurrencyCode(),toCurrency.getCurrencyCode(),false);
        }
        return result;
    }
    
    /**
     * Convert <code>value</code> from <code>fromCurrency</code> to <code>toCurrency</code>.
     * Exchange rate is provided by exchange rate provider which was specified when Convertor was created.
     * <p>
     *  This method is using only exchange rate from->to and if not found, it is trying to use reverted excange rate to->from.
     * 
     * @param fromCurrency Source currency to convert from.
     * @param toCurrency Target currency to convert to.
     * @param value Value in source currency which should be converted.
     * @return Return conversion result.
     * @since version2
     * @throws ConversionNotSupportedException If conversion from <code>fromCurrency</code> to <code>toCurrency</code> 
     *         is not supported and neither conversion from <code>toCurrency</code> to <code>fromCurrency</code> is not supported.
     */
    public ConversionResult convertWithReversibleRates(ConvertorCurrency fromCurrency, ConvertorCurrency toCurrency, BigDecimal value) {
        ConversionResult result = convertValue(fromCurrency, toCurrency, value, false,true);
        if (result==null) {
            //throw new ConversionNotSupportedException("Neither onversion nor reverted conversion from " + fromCurrency + " to " + toCurrency + "  is not supported,");
            throw new ConversionNotSupportedException(fromCurrency.getCurrencyCode(),toCurrency.getCurrencyCode(),true);
        }
        return result;
    }
    
}
